/*
 * Aserver client test - Asyncronous send and receive
 *
 * There is a potential for deadlock if we have two Aserver connections.
 * You must never wait on one connection if another may receive data.
 * Here we have a hidden second Aserver connection to ISERV.
 * A process using ISERV (print etc) will wait till it gets the ISERV status reply.
 * This ISERV reply may be held in Aserver behind a packet destined for
 * our other connection. If the receive process is waiting for the print
 * to complete we are in a deadlock.
 */

#include <newdef.h>
#include <assert.h>
#include <process.h>
#include <channel.h>
#include <misc.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "aserver.h"

/* Aserver test maximum message length */
#define ASTEST_MAXMESSAGE 100

/* same as assert, but excuted even when not debug compile */
#ifdef NDEBUG
#       define check(e) ((e)!=0 ? (void)0 : (void)abort())
#else
#       define check(e) assert((e)!=0)
#endif


PRIVATE void aserver_error (asrc error)
/* Aserver error handling callback */
{
	assert (error != ASRC_SUCCESS);
	/* print it out if i can ! */
	puts (as_asrc_to_str(error));
	getchar();
	/* not much else we can do */
	abort();
}


PRIVATE void pfSend (Process *p, BOOL bSync, Channel* Isync)
{
	int i;
	char TestMessage[ASTEST_MAXMESSAGE];
	p = p;

	/* Form test message */
	for (i = 0; TRUE ; i++) {
		sprintf (TestMessage, "Testing %d", i);

		aserver_send (TestMessage, strlen(TestMessage));
		if (bSync)
			(void)ChanInChar (Isync);
	}
}

PRIVATE void pfReceive (Process *p, BOOL bSync, Channel* Osync, Channel *IprintReady, Channel *Oprint)
/*
 * No ISERV (printf etc) in here. It will deadlock if we are not ready
 * to aserver_receive
 */
{
	unsigned long dl;
	char ReplyMessage[ASTEST_MAXMESSAGE];
	BOOL done = FALSE;
	p = p;

	do {
		dl = sizeof(ReplyMessage);
		
		aserver_receive (ReplyMessage, &dl);
		
		if (bSync)
			ChanOutChar (Osync,1);
			
		if ((ProcSkipAlt(IprintReady) == 0) || done) {
			(void) ChanInInt (IprintReady);
			ChanOutInt (Oprint, (int)dl);
			ChanOut (Oprint, ReplyMessage, (int)dl);
		}
	} while (!done);
}


PRIVATE void pfPrintReceive (Process *p, Channel *OprintReady, Channel *Iprint)
{
	int dl;
	char ReplyMessage[ASTEST_MAXMESSAGE+1];
	BOOL done = FALSE;
	p = p;
	
	do {
		ChanOutInt (OprintReady, 1);

		dl = ChanInInt (Iprint);		
		ChanIn (Iprint, ReplyMessage, dl);
		
		ReplyMessage[dl] = '\0';
		puts (ReplyMessage);
	} while (!done);
}


int main()
{
	/* Process Configuration */
	int c = 3;
	Channel* ciAserver = (Channel*) get_param (c++);
	Channel* coAserver = (Channel*) get_param (c++);
	char* strServerService = (char*) get_param (c++);
	BOOL bSync = *(BOOL*) get_param (c++);

	/* Locals*/
	Channel *CprintReady, *Cprint, *Csync;
	Process *Psend, *Preceive, *PreceivePrint;
 
	check ( CprintReady = ChanAlloc() );
	check ( Cprint = ChanAlloc() );
	check ( Csync = ChanAlloc() );
	check ( Psend = ProcAlloc (pfSend, 0, 2, bSync, Csync) );
	check ( Preceive = ProcAlloc (pfReceive, 0, 4, bSync, Csync, CprintReady, Cprint) );
	check ( PreceivePrint = ProcAlloc (pfPrintReceive, 0, 2, CprintReady, Cprint) );


	/* Make configured connection to Server */
	if (aserver_start (ciAserver, coAserver, strServerService, aserver_error) != ASRC_SUCCESS) {
		fprintf (stderr, "testcli: Couldn't start Aserver connection to \"%s\"\n", strServerService);
		exit (EXIT_FAILURE);
	}

	fprintf (stderr, "testcli: Started Aserver connection to \"%s\"\n", strServerService);

	/* Start Send/Receive Processes */
	ProcPar (Psend, Preceive, PreceivePrint, NULL);
	
	/* Finish with Server */
	fprintf (stderr, "testcli: About to finish\n");
	aserver_finish ();

	exit (EXIT_SUCCESS);
}

