/*
 * AServer Support:
 *
 * Implements a multiple connection interface to the ASERVER protocol.
 * Only a single instance exists which encapsulates a ASERVER access point.
 *
 * Abstract class with some Pure Virtual functions that complete the processing
 * of received messages and provide error handling.
 *
 * Received packets are buffered into a common FIFO buffer from ASERVER callbacks.
 * Packets are immediately processed and data packets
 * are then transfered into an instance's message buffer.
 *
 * Connection requests are automaticly accepted until the maximum connection limit is reached.
 * All connections are aborted when the instance is distructed
 * 
 * Overflow of the common FIFO or an instance message buffer
 * generates a protocol error message and throws a user exception.
 *
 * Aserver receives data automatically from inside any windows message loop.
 * This also means Modal Dialog box's etc, even sometimes inside SendMessage.
 * Take care as some functions will be reentered in this case.
 *
 * s_HoldAsync - STATIC MEMBER
 *  Allows the asyncronous behaviour of ASERVER to be disables/enabled. This
 *  can only be perform for the whole access point
 *
 *
 * SendMessage - MEMBER
 *	Send a message. Call it at anytime but be careful as messages can be received
 *	asyncronously when it is being called.
 *
 * ProtocolFail - PURE VIRTUAL
 *  Reports a fatal communcations protocol error
 *  Never reenters
 *  
 * ConnectionInfo - PURE VIRTUAL
 *  Reports connections disconnections and other none data information
 *  In connection requests the service name and parameters are supplied
 *  May be reentered
 *
 * RecMessage - PURE VIRTUAL
 *	Provides processing of received messages.
 * 	Returns TRUE to indicate completed received message processing.
 * 	Otherwise acknowledge completion of processing later with RecMessageCompleted.
 * 	Until RecMessage return TRUE or RecMessageCompleted is called the received
 *	message buffer is locked and received packets will be buffered in the common FIFO.
 * 
 */

#include "stdlib.h"
#include "stdio.h"
#include "string.h"
#include "assert.h"

#include "aserver.h"


// Class variables
CAserver*	CAserver::s_active = NULL;
BOOL		CAserver::s_SIGINT = FALSE;
BOOL		CAserver::s_criticalProcess = FALSE;
BOOL		CAserver::s_criticalFIFO = FALSE;
BOOL		CAserver::s_criticalSend = FALSE;

/*
 * Aserver common protocol
 * Packet FIFO, initialised by CAserver constructor.
 * Somewhat protected by sanity check of s_criticalFIFO, this should never be
 * violated as neither ASERVER callbacks or Windows messges are preemptive.
 */
unsigned	CAserver::s_nR;
unsigned	CAserver::s_nW;
unsigned	CAserver::s_nSize;
APacket*	CAserver::s_pFifo;


asrc CAserver::s_AddPacket (int ap_num, packet_hdr hdr, unsigned char* data)
/* Add a packet into common FIFO
 * Used as standard asyncronous callback function, therefore no "const" in decleration */
{
	// Enter FIFO Critical Section
	assert (!s_criticalFIFO);
	s_criticalFIFO = TRUE;
	
	asrc res = ASRC_SUCCESS;

	s_pFifo[s_nW].ap_num = ap_num;
    memcpy(s_pFifo[s_nW].hdr, hdr, sizeof(packet_hdr));
    memcpy(s_pFifo[s_nW].data, data, (int) as_hdr_get_len(hdr));

	s_nW++; if (s_nW == s_nSize) s_nW = 0;
	if (s_nW == s_nR)	{		// FIFO full eventually generate a protocol error
		res = ASRC_OUT_OF_MEMORY;
	}

	s_criticalFIFO = FALSE;
	return res;
}

asrc CAserver::s_GetPacket (int* ap_num, packet_hdr hdr, unsigned char* data)
/* Get a packet from common FIFO */
{
	// Enter FIFO Critical Section
	assert (!s_criticalFIFO);
	s_criticalFIFO = TRUE;
	
	asrc res = ASRC_SUCCESS;

	if (s_nR == s_nW) {
		res = ASRC_FAILED;
		assert (FALSE);			// Must be a internal error
	}

	*ap_num = s_pFifo[s_nR].ap_num;
    memcpy(hdr, s_pFifo[s_nR].hdr, sizeof(packet_hdr));
    memcpy(data, s_pFifo[s_nR].data, (int) as_hdr_get_len(hdr));

	s_nR++; if (s_nR == s_nSize) s_nR = 0;

	s_criticalFIFO = FALSE;
	return res;
}


BOOL CAserver::s_Empty ()
{
	// Enter FIFO Critical Section
	assert (!s_criticalFIFO);
	return s_nR == s_nW;
}


/*
 * Aserver common protocol
 */
void CAserver::s_AsyncHold ()
{
	ass_set_cb (NULL);
}

void CAserver::s_ProtocolFail (const char* message, asrc result, packet_hdr header)
{
	char error[500];
	char headerText[200] = "";
	
    // Disable Async callback ASAP so we cannot renter
	s_AsyncHold ();

	// Must be active
	assert (s_active);
	
	// Format an error message                                
	if (header != NULL)
		as_hdr_to_str(headerText, header);

	if (result == ASRC_SUCCESS)	{
		if (header == NULL)
			sprintf (error, "ASERVER: %s", message);
		else
			sprintf (error, "ASERVER: %s\n%s", message, headerText);
	}
	else
		sprintf (error, "ASERVER: %s\nresult=%s %s", message, as_asrc_to_str(result), headerText);

	s_active->ProtocolFail (error);
}


void CAserver::s_ProcessBuffer ()
/*
 * Process any buffered packets if the client is ready for a new message.
 * Packets must be process "in-order", i.e. interpreted in the order they are received.
 * The main processing of packets is therefore protected by a sanity check using s_criticalProcess.
 * When the client is busy no further packets can be processed. They will
 * remain buffered until the client calls RecMessageCompleted.
 * N.B. The processing itself may result in more packets being buffered
 */
{
	asrc		res;
	packet_hdr	hdr;
	unsigned char data[MAX_PACKET_SIZE];
	int apNum;
	
	// We must have an active client    
	assert (s_active != NULL);

	// First check for ASERVER SIGINT
	if (s_SIGINT) {
		asrc res;

	    // Disable Async callback ASAP
	    s_AsyncHold ();

		res = as_apfinish ();
		if (res != ASRC_SUCCESS)
			s_ProtocolFail("Shutdown, finish unsucessful", res);
        else
			s_active->ConnectionInfo (0, 0, NULL, TRUE, NULL, NULL);
		s_active = NULL;
	
		return;
	}

	// Delay processing while client is busy
	if (s_active->m_bClientLock)
		return;

	// Check not reentering critical section
	// Prevents me from accidently recursing, we can easily happen if
	// from the Windows callback function
	assert (!s_criticalProcess);


	// Iterate through buffered packets unless the Client is not ready
	while (s_active && !s_active->m_bClientLock && !s_Empty()) {
		BOOL bReportConnectionReq = FALSE;
		BOOL bReportConnectionInfo = FALSE;
		int c_num;
		
		// Enter critical section, packet processing must be completed to maintain
		// in-order processing.
		s_criticalProcess = TRUE;

		// Get buffered packets
		res = s_GetPacket (&apNum , hdr, data);
		res = as_translate_pkt (res, apNum, hdr, data, s_AddPacket);

		// Useful parts
		c_num = as_hdr_get_conn_num (hdr);

		switch (res) {        

			case ASRC_SUCCESS: {
				AConnection* pConnection = &(s_active->m_pConnects[c_num]);
				int type = as_hdr_get_type (hdr);    
						
				switch (type & ASPH_TYPE_MASK) {
					case ASPT_CONNECT_REQ: {	// Connection Packet - process connection 
						res = ass_processconnect (apNum, hdr, data, &c_num, ASRC_SUCCESS, s_AddPacket);
						if (res != ASRC_SUCCESS)
							s_ProtocolFail ("Connection error", res, hdr);
						else {	
							pConnection->m_bConnected = TRUE;
							s_active->m_nConnect++;
							if (s_active->m_nConnect > s_active->m_nMaxConnect)
								s_ProtocolFail ("Too many connections");
                	        else 
								bReportConnectionReq = TRUE;
						}
					}
					break;          
							
					case ASPT_DATA: {        	// Data Packet - add data to received message
						UINT len = as_hdr_get_len (hdr);
						if (pConnection->m_nRecNum + len > pConnection->m_nRecSize)
							s_ProtocolFail ("Message too long for client buffer");
						else {							
							memcpy ((LPBYTE)pConnection->m_pRecData + pConnection->m_nRecNum, data, len);
							pConnection->m_nRecNum += len;
							
							if (ASPH_IS_EOM(type)) {		// Received message complete
								s_active->m_bClientLock = TRUE;
								s_active->m_bClientLock = !s_active->RecMessage (c_num, pConnection->m_pRecData, pConnection->m_nRecNum);
								pConnection->m_nRecNum = 0;
							}
						}
					}
					break;
				}/*switch packet type*/
						
							
			}/*ASRC_SUCCESS*/
			break;

			case ASRC_GOT_DISCONNECT_REQ:
				if (s_active->m_nConnect == 0)
					s_ProtocolFail ("Too many disconnections");
				else
					s_active->m_nConnect--;

				bReportConnectionInfo = TRUE;
			break;
	
			/* Debugger Control Messages */
			case ASRC_GOT_ABORT_REQ:
			case ASRC_TX_ERROR:
			case ASRC_RESTART:
			case ASRC_TERMINATE:
				bReportConnectionInfo = TRUE;
			break;
             
			default: {
				s_ProtocolFail ("Untranslated packet", res, hdr);
			}
			break;
		}/*switch translate result*/;
				
		// Leave critical section packet has completed its processing
		s_criticalProcess = FALSE;

		// Report connection request or other information
		assert (!(bReportConnectionReq && bReportConnectionInfo));

		if (bReportConnectionReq) {
			char *service, *serviceParams;
			as_decode_conn_req (data, &service, &serviceParams);
			
			s_active->ConnectionInfo (c_num, res, hdr, FALSE, service, serviceParams);
		}
		if (bReportConnectionInfo)
			s_active->ConnectionInfo (c_num, res, hdr, FALSE, NULL, NULL);

	}/* while more to do */

}


void CAserver::s_Callback (short aserverEvent)
/*
 * Aserver callback. Somehow called automatically from windows message loop.
 * This also means Modal Dialog box's etc, even sometimes inside as_send_packet!
 */
{
	switch (aserverEvent) {
		case ASCB_RECPKT: {				// Packet ready to read
			asrc		res;
			packet_hdr	hdr;
			unsigned char data[MAX_PACKET_SIZE];
			int apNum = 0;
			res = as_rec_packet (apNum, hdr, data);
			if (res != ASRC_SUCCESS) {
				/* Ignore the generic error as we get it on SIGINT */
				if (res != ASRC_FAILED)
					s_ProtocolFail ("Receive failure", res, NULL);
			}
			else if (s_AddPacket (apNum, hdr, data) != ASRC_SUCCESS)
				s_ProtocolFail ("Internal common receive buffer overflow");
			else	// Start processing this packet
				s_ProcessBuffer ();
		}/*case ASCB_RECPKT*/
		break;

		
		case ASCB_SIGINT: {  
			// Not sure if SIGINT also possible inside as_send_packet
			// Just incase delay processing if inside SendMessage and
			// let the subsequent call to s_ProcessBuffer handle it.
			s_SIGINT = TRUE;
			if (!s_criticalSend)
				s_ProcessBuffer ();
		}
		break;
		
		default:
			s_ProtocolFail("Unknown ASERVER callback event");
		break;		
	}/*switch aserverEvent*/
}


/*
 * CAserver connection instance
 */


CAserver::CAserver (unsigned nMaxConnect, unsigned nMaxRecMessageSize, unsigned nRecCommonFIFO)
/*
 * Initialize the ASERVER access point instance with a given number of connections points.
 * With 0 connection points ASERVER is not initialised and subsequent arguments are ignored
 * Messages of length nMaxRecMessageSize can be received. Larger messages will generate
 * a protocol error as they are assembled.
 * nRecCommonFIFO determines the number of packets in the Common FIFO buffer.
 */
{
	asrc res = ASRC_SUCCESS;                      
	
	// Only one instance of CAserver
	assert (s_active == NULL);

	// No connecttions yet
	m_nConnect = 0;
	m_nMaxConnect = nMaxConnect;

	if (nMaxConnect > 0) {
		s_active = this;

		// Connections provide message receive buffers
		m_pConnects = new AConnection[nMaxConnect];
		for (unsigned i = 0; i < nMaxConnect; i++)
			m_pConnects[i].ArrayInit (nMaxRecMessageSize);
	
		// Message buffer processing
		m_bClientLock = FALSE;
			
		// Common Receive FIFO
		s_nR = s_nW = 0;
		s_nSize = nRecCommonFIFO;
		s_pFifo = new APacket[s_nSize];
		
	    // Register the Async callback now. It must be live for the first packet
		ass_set_cb (s_Callback);
	
		// Initialize ASERVER access point and connection table
		res = as_apstart();
		if (res != ASRC_SUCCESS) {
			s_ProtocolFail ("Failed to initialise access point", res);
		}
		else {
			m_pConnTable = new conn_id[nMaxConnect];
			res = as_setconntable (nMaxConnect, m_pConnTable);
	
			if (res != ASRC_SUCCESS) {                                                         
				s_ProtocolFail ("Failed to as_setconntable", res);
			}
		}
	}
	else {
		// No connections, no ASERVER
		m_pConnects = NULL;
		m_pConnTable = NULL;
		s_pFifo = NULL;
	}
}


CAserver::~CAserver ()
/*
 * Close down the Aserver access point
 * Abort any remaining connections
 * It is nolonger possible to issue ProtocolFail as any derived object no longer exists
 */
{
   	asrc res;

	if (s_active) {
	    // Disable Async so we can procces abort directly
    	s_AsyncHold ();
    	
        // Abort connections if made
        for (unsigned i = 0; i < m_nMaxConnect; i++) {
        	if (m_pConnects[i].m_bConnected) {
				res = ass_sendabort(i, NULL);
				/* Ignore result it is too late */
			}
		}

	    // Finish
		res = as_apfinish ();
		/* Ignore result it is too late */
		s_active = NULL;
	}

	if (s_pFifo) delete [] s_pFifo;
	if (m_pConnTable) delete [] m_pConnTable;
	if (m_pConnects) delete [] m_pConnects;
}


void CAserver::RecMessageCompleted ()
/* Acknowledge delayed completion of processing of received message.
 * Call ONLY once, ONLY after returning from RecMessage with FALSE.
 * Received message buffer is unlocked. Packet processing is restarted and RecMessage
 * may now be called at any time.
 * Note: Make sure you really have completed all message processing before calling.
 *       In particular if you use SendMessage in reply only call after
 *       it has completed. ASERVER receives packets and sends packets asyncronously.
 *       A new message MAY have been received while inside SendMessage and RecMessage
 *       function will be rentered before SendMessage completes.
 */
{
	assert (s_active == this);
	assert (m_bClientLock);	// Only called in response to OnRecMessage
	
	// Restart packet processing
	m_bClientLock = FALSE;
	s_ProcessBuffer ();
}


void CAserver::SendMessage (int c_num, const void* data, unsigned size)
/* Send an Aserver message.
 *                                 
 * ASERVER allows asyncronous sending and reception of packets. To allow
 * this it seems that as_send_packet (which we use) allows some windows message processing.
 * Therefore application sometimes receives windows messages before SendMessage completes.
 *
 * Packets may be recevied while executing SendMessage, see RecMessageComplete notes.
 * Protected by s_criticalSend to prevent packets of different messges being merged.
 */
{
	asrc res;                                   

	// Enter Send Critical Section, ensure message is complete before next is sent
	assert (!s_criticalSend);
	s_criticalSend = TRUE;
	
	res = as_sendmessage (c_num, size, (unsigned char*)data, s_AddPacket);

	s_criticalSend = FALSE;

	if (s_active) 	/* May have finished in the send */
	{
		if (res != ASRC_SUCCESS)
			s_ProtocolFail("Send message failure", res);
	    else	// Packets may be buffered by aserver callback
			s_ProcessBuffer ();
	}
}
